/*! \file CMT1.h
	\brief	Contains the CMT Level 1 interfaces

	This level contains OS-dependent implementations of Serial and File 
	communication, as well as some other functionality like 'sleep'.

	\date 2006-04-12	Job Mulder, Created
*/

#ifndef CMT1_H_2006_04_12
#define CMT1_H_2006_04_12

//! The default namespace of all Xsens software since 2006.
namespace xsens {

//////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Constants  ///////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief	The return value of virtually all CMT routines.
	
	This value should always be checked, at least against CMTResult::OK.
*/
enum CMTResult {
	OK		= 0,		//!< Operation was performed succesfully
	ERROR,				//!< A generic error occurred
	TIMEOUT,			//!< A timeout occurred
	NOT_FOUND,			//!< The value was not found
	NOT_AVAILABLE,		//!< The requested data is not available

	__RESULT_COUNT		//!< The number of items in the CMTResult enum
};

/*! \brief	A textual representation of each CMTResult.

	This list is used by the resultAsText function.
*/
extern const char CMTResultText[CMT__COUNT+1][];
/*
The CMT1.cpp file will contain something like this:

const char CMTResultText[CMT__COUNT+1][] = {
	"No error",
	"Unknown error",
	"Timeout",
	"Value not found",

	"!!Invalid result code!!"
};
*/

/* \brief Retrieve a string corresponding to the given result code.

	This function uses the CMTResultText list to return a string with the relevant result 
	code in a textual format. If the supplied result code is invalid (>=__RESUL_COUNT) the
	"!!Invalid result code!!" string is returned.
*/
const char[] resultAsText(CMTResult result);

//! The default size of the CMT1s read buffer in bytes
#define CMT_DEFAULT_READ_BUFFER_SIZE		512
//! The default baud rate of the CMT1s serial communication
#define CMT_DEFAULT_BAUD_RATE				115200
//! The default timeout value for blocking CMT1s operations in ms
#define CMT_DEFAULT_TIMEOUT					10

enum ControlLine {
	CONTROL_RI		= ...,
	CONTROL_CTS		= ...,
	CONTROL_RTS		= ...,
	CONTROL_...		= ...
};

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// CMT1s  /////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief The low-level serial communication class.
*/
class CMT1s {
private:
	//! This object cannot be copied, so this function is not implemented.
	CMT1s(const CMT1s& ref);

protected:
	unsigned char m_readBuffer*;	//!< The buffer used to store received data
	unsigned int m_readBufferSize;	//!< The size of m_readBuffer in bytes
	unsigned int m_readPos;			//!< The current read position in m_readBuffer
	unsigned int m_readCount;		//!< The current count of unread bytes in m_readBuffer

	unsigned char m_port;			//!< The opened COM port
	bool m_isOpen;					//!< Indicates if the port is open or not

	CMTResult m_lastResult;			//!< The last result of an operation

	/*! The default timeout value to use during blocking operations.
		A value of 0 means that all operations become non-blocking.
	*/
	unsigned int m_timeout;
public:
	//! Default constructor, initializes all members to their default values.
	CMT1s(unsigned int readBufSize = CMT_DEFAULT_READ_BUFFER_SIZE);

	//! Destructor, de-initializes, frees memory allocated for buffers, etc.
	~CMT1s();

	//! \brief Close the serial communication port.
	CMTResult close (void);

	/*! \brief Flush all data to be transmitted / received.
		This function tries to send and receive any remaining data immediately 
		and does not return until the buffers are empty.
	*/
	CMTResult flushData (void);

	//! \brief Retrieve the port number that was last succesfully opened.
	CMTResult getName (unsigned char& portNumber) const;

	//! \brief Return whether the communication port is open or not.
	bool isOpen (void) const;

	/*! \brief Open a communication channel to the given COM port number.
	
		The port is automatically initialized to the given baud rate.
		If the baudrate is set to 0, the baud rate is automatically detected. If possible.
	*/
	CMTResult open (const unsigned char portNumber,
					const unsigned long baudRate = CMT_DEFAULT_BAUD_RATE);

	/*! \brief Read data from the serial port and put it into the data buffer.

		This function reads as much data as possible from the com port (non-blocking) and 
		put as much data as will fit into the data buffer. Any excess data is stored in 
		the \c m_readBuffer member variable. If there was enough data in m_readBuffer to 
		fulfill the request, the data parameter is first filled and the port is polled 
		afterwards.
		\param maxLength	The maximum amount of data read.
		\param data			Pointer to a buffer that will store the received data.
		\param length		The number of bytes placed into \c data.
	*/
	CMTResult readData (const unsigned long maxLength, unsigned char* data,
														unsigned long& length);

	/*! \brief Set the sizes of the inbound and outbound data buffers.
	
		The function resizes the m_readDatabuffer buffer.
		Any excess data is discarded, but existing data is retained.
	*/
	CMTResult setReadBufferSize (const unsigned long readBufSize);

	/*! \brief Set the default timeout value to use in blocking operations.

		This function sets the value of m_timeout. There is no infinity value. The value 0 
		means that all blocking operations now become polling (non-blocking) operations. 
		If the value is set to or from 0, the low-level serial port settings may be 
		changed in addition to the m_timeout value.
	*/
	CMTResult setTimeout (const unsigned long ms = CMT_DEFAULT_TIMEOUT);

	/*! \brief Wait for data to arrive or a timeout to occur.

		The function waits until some data is available or until a timeout occurs. The 
		function returns success if data is available or CMTResult::TIMEOUT if a timeout 
		occurred.
	*/
	CMTResult waitForData (const unsigned long timeout) const;

	/*! \brief Write the data to the serial port.

		The function writes the given data to the connected COM port.
		The default timeout is respected in this operation.
	*/
	CMTResult writeData (const unsigned long length,  const unsigned char* data);

	//! Return the error code of the last operation.
	CMTResult getLastError(void);

	/*! \brief Manipulate the Serial control lines

		The function manipulates the serial control lines that are indicated by the 
		mask parameter.
		\param mask		Indicates which lines are to be manipulated and which should be
						left alone.
		\param state	Contains the new state of the control lines.
	*/
	CMTResult escape (const ControlLine mask, const ControlLine state);
};

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// CMT1f  /////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief The low-level file communication class.
*/
class CMT1f {
private:
	//! This object cannot be copied, so this function is not implemented.
	CMT1f(const CMT1f& ref);

protected:
	unsigned int m_readPos;			//!< The last read position in the file
	unsigned int m_writePos;		//!< The last write position in the file

	CMTResult m_lastResult;			//!< The last result of an operation

	/*! \brief Indicates whether the last operation was a read or write operation.

		This value is used to check whether or not a seek is required to perform a 
		requested read or write operation.
	*/
	bool m_lastOpWasRead;
	bool m_isOpen;			//!< Indicates whether the file is open or not.
	
	//! Contains the name of the file that was last succesfully opened.
	char m_filename[256];

public:
	//! Default constructor, initializes all members to their default values.
	CMT1f();

	//! Destructor.
	~CMT1f();

	//! \brief Close the file.
	CMTResult close (void);

	/*! \brief Flush all data to be written.
		This function writes any remaining data immediately and does not return
		until this is done.
	*/
	CMTResult flushData (void);

	/*! \brief Retrieve the filename that was last succesfully opened.

		\param filename	A buffer for storing the filename. The buffer should be able 
				to hold the filename. A safe size is to make it at least 256 bytes.
	*/
	CMTResult getName (unsigned char* filename) const;

	/*! \brief Return the current read position.
	*/
	unsigned long getReadPos(void) const;

	/*! \brief Return the current write position.
	*/
	unsigned long getWritePos(void) const;

	//! \brief Return whether the file is open or not.
	bool isOpen (void) const;

	/*! \brief Open a file.
	*/
	CMTResult open (const char* filename);

	/*! \brief Read data from the file and put it into the data buffer.

		This function reads exactly the number of bytes as requested from the file.
		\param maxLength	The amount of data that will be read.
		\param data			Pointer to a buffer that will store the read data.
	*/
	CMTResult readData (const unsigned long maxLength, unsigned char* data);

	/*! \brief Read data from the file and put it into the data buffer.

		This function reads upp to the number of bytes as requested from the file.
		The function will also stop if the given terminator character is encountered.
		The terminator is included in the output buffer.
		\param maxLength	The amount of data that will be read.
		\param data			Pointer to a buffer that will store the read data.
		\param terminator	A character that will end the read operation if encountered.
		\param length		The actual number of bytes read, including the terminator
							character, if encountered.
	*/
	CMTResult readData (const unsigned long maxLength, const char terminator,
						unsigned char* data, unsigned long& length);

	/*! \brief Set the new absolute read position

		The read position is checked against the filesize before committing.
	*/
	CMTResult setReadPos (const unsigned long pos);

	/*! \brief Set the new absolute write position

		The write position is checked against the filesize before committing.
	*/
	CMTResult setWritePos (const unsigned long pos);

	/*! \brief Find a string of bytes in the file

		The function searches from the current read position until the given \c needle is 
		found. If the needle is not found, CMTResult::NOT_FOUND is returned. The function
		will update the seek position to the first character after the found needle.
		\param needle		The byte string to find.
		\param needleLength	The length of the byte string.
		\param pos			Out: The position where the needle was found. This will point
							to the first character of the found needle.
	*/
	CMTResult find (const unsigned char* needle, const unsigned long needleLength,
															unsigned long& pos);

	/*! \brief Write data to the file.

		The function writes the given data to the file at the current write position.
	*/
	CMTResult writeData (const unsigned long length,  const char* data);

	//! Return the error code of the last operation.
	CMTResult getLastError(void);
};

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Other  functions ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief A platform-independent sleep routine.

	Time is measured in ms. The function will not return until the specified
	number of ms have passed.
*/
void msleep(unsigned int ms);

}	// end of xsens namespace

#endif
