/*! \file CMT2.h
	\brief	Contains the CMT Level 2 interfaces

	This level contains the message- and log-file interfaces of CMT.
	All code should be OS-independent.

	\date 2006-04-13	Job Mulder, Created
*/

#ifndef CMT2_H_2006_04_13
#define CMT2_H_2006_04_13

#include "CMT1.h"

namespace xsens {

//////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Constants  ///////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////


//! Predefined BusIds
enum BusIds {
	BID_BROADCAST	= 0,
	BID_MASTER 		= 255
};


//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Support  classes ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief Class for storing a single message.

	The 
*/
class Message {
protected:
	unsigned long m_createdLength;

	union _mbh {
		unsigned char* m_buffer;
	
		struct MessageHeader {
			unsigned char m_preamble;
			unsigned char m_busId;
			unsigned char m_messageId;
			unsigned char m_length;
			union _mdl {
				struct _mextd {
					unsigned short m_extendedLength;
					unsigned char m_extendedData[0];
				};
				unsigned char m_data[0];
			};
		}* m_header;	//!< The message header is the data buffer with interpretation
	};
	
	//! Internal function to get the start of the data buffer.
	unsigned char* getDataStart(void);
	
public:
	/*! \brief Create a Message object with the given data length.
	
		The function allocates enough memory to hold an entire message with the given 
		data length.
		\param length	The length of the data in the message. This value is stored in 
						\c m_createdLength as well as \c m_length or \c m_extendedLength.
	*/
	Message(const unsigned short length);

	/*! \brief Create a Message object with the given data length and message Id.
	
		The function allocates enough memory to hold an entire message with the given 
		data length.
		\param length	The length of the data in the message. This value is stored in 
						\c m_createdLength as well as \c m_length or \c m_extendedLength.
		\param msgId	The message Id that will be assigend to the m_messageId field.
	*/
	Message(const unsigned short length, const unsigned char msgId);

	/*! \brief Destroy a Message object.
	
		The function deallocates the memory associated with the Message.
	*/
	~Message();

	/*! \brief Read the entire message from the given source string
	
		This is done through a simple memory copy. The number of bytes copied is \c
		m_createdLength. After the copy, the length reported in the message is checked
		against the \c m_createdLength member.
	*/
	CMTResult loadFromString(const unsigned char* source);
	
	/*! \brief Return the start of the message buffer.
	
		The function returns the address of the \c m_preamble member.
	*/
	const unsigned char* getMessageStart(void);
	
	/*! \brief Return the start and length of the message buffer.
	
		The function returns the total size of the message, including the checksum. This
		is in effect the number of bytes that would be transferred if the message were to
		be sent over a communications channel.
	*/
	const unsigned int getMessageSize(void);
	
	//! Return the current value of the m_busId field.
	unsigned char getBusId(void);
	
	//! Set the new value of the m_busId field and update the checksum.
	void setBusId(unsigned char busId);
	
	//! Return the current value of the m_messageId field.
	unsigned char getMessageId(void);
	
	//! Set the new value of the m_messageId field and update the checksum.
	void setMessageId(unsigned char msgId);
	
	//! Return the length of the message data.
	unsigned short getLength(void);
	
	/*! \brief Return the current value of the data as an unsigned byte (8 bits).
	
		\param offset An optional offset in the data buffer from where to start reading.
	*/
	unsigned char getDataByte(const unsigned long offset=0);
	
	/*! \brief Return the current value of the data as an unsigned short (16 bits).
	
		\param offset An optional offset in the data buffer from where to start reading.
	*/
	unsigned short getDataShort(const unsigned long offset=0);
	
	/*! \brief Return the current value of the data as an unsigned long (32 bits).
	
		\param offset An optional offset in the data buffer from where to start reading.
	*/
	unsigned long getDataLong(const unsigned long offset=0);
	
	/*! \brief Return the current value of the data as a float (32 bits).
	
		\param offset An optional offset in the data buffer from where to start reading.
	*/
	float getDataFloat(const unsigned long offset=0);
	
	/*! \brief Return a pointer to the data buffer.
	
		\param offset An optional offset in the data buffer from where to start reading.
	*/
	unsigned char* getDataBuffer(const unsigned long offset=0);

	/*! \brief Write an unsigned byte (8 bits) into the data buffer.
	
		\param data		The data to write to the buffer.
		\param offset	An optional offset in the data buffer from where to start writing.
	*/
	CMTResult setDataByte(const unsigned char data, const unsigned long offset=0);
	
	/*! \brief Write an unsigned short (16 bits) into the data buffer.
	
		\param data		The data to write to the buffer.
		\param offset An optional offset in the data buffer from where to start writing.
	*/
	CMTResult setDataShort(const unsigned short data, const unsigned long offset=0);
	
	/*! \brief Write an unsigned long (32 bits) into the data buffer.
	
		\param data		The data to write to the buffer.
		\param offset An optional offset in the data buffer from where to start writing.
	*/
	CMTResult setDataLong(const unsigned long data, const unsigned long offset=0);
	
	/*! \brief Write a float (32 bits) into the data buffer.
	
		\param data		The data to write to the buffer.
		\param offset An optional offset in the data buffer from where to start writing.
	*/
	CMTResult setDataFloat(const float data, const unsigned long offset=0);
	
	/*! \brief Write a string of bytes into the data buffer.

		\param data		The data to write to the buffer.
		\param offset	An optional offset in the data buffer from where to start writing.
		\param count	An optional number of bytes to write, if set to 0 (not set), as
						many bytes as will fit into the buffer from the given point will 
						be written.
	*/
	CMTResult setDataBuffer(const unsigned char* data, const unsigned long offset=0,
							const unsigned long count = 0);

	//! Compute the checksum and compare it with the stored checksum. Equal is ok.
	bool isChecksumOk(void);
	
	/*! \brief Compute the checksum field and fill it.
	
		The checksum field should normally be correct at all times, but if you have
		somehow managed to mess it up, this function can be used to recompute it.
	*/
	CMTResult computeChecksum(void);
};


/* different alignment commands for gcc / MSVS, the structure needs to be 2-byte aligned
	because the deviceId field is 4 bytes long on offset 98.
*/
#ifdef _MSC_VER
	#pragma pack(push, 2)
#endif
//! Structure containing a full device configuration as returned by the ReqConfig message.
struct DeviceConfiguration {
	unsigned long	masterDeviceId;
	unsigned short	samplingPeriod;
	unsigned short	outputSkipFactor;
	unsigned short	syncinMode;
	unsigned short	syncinSkipFactor;
	unsigned long	syncinOffset;
	unsigned char	date[8];
	unsigned char	time[8];
	unsigned char	reservedForHost[32];
	unsigned char	reservedForClient[32];
	unsigned short	numberOfDevices;
	unsigned long	deviceId;
	unsigned short	dataLength;
	unsigned short	outputMode;
	unsigned long	outputSettings;
	unsigned char	reserved[8];
}
// different alignment commands for gcc / MSVS
#ifdef _MSC_VER
	;
	#pragma pack(pop)
#else
	__attribute__((__packed__));
#endif

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// CMT2s  /////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief Mid-level serial communication class.

	The class uses CMT level 1, but does not inherit from it. If software needs to access 
	the level 1 component, it needs to be done through the cmt1s() function.
*/
class CMT2s {
private:
	//! This object cannot be copied, so this function is not implemented.
	CMT2s(const CMT2s& ref);

protected:
	//! The CMT level 1 object that this class operates on
	CMT1s m_cmt1s;

	/*! \brief Compute the checksum of the given byte string.
	*/
	static CMTResult computeChecksum(unsigned char* buffer, unsigned long length);

	CMTResult m_lastResult;			//!< The last result of an operation
	
public:
	//! Default constructor, initializes all members to their default values.
	CMT2s();

	//! Destructor, de-initializes, frees memory allocated for buffers, etc.
	~CMT2s();

	/*! \brief Close the serial communication port.
	
		Buffered data remains available until the object is destroyed or until the open()
		function is called again.
	*/
	CMTResult close (void);

	/*! \brief Return a reference to the embedded CMT1s object.
	
		Any manipulation of the object should be done through CMT2s. CMT2s integrity is 
		not guaranteed if the CMT1s object is manipulated directly.
	*/
	CMT1s* cmt1s(void);

	//! Return the error code of the last operation.
	CMTResult getLastError(void);

	/*! \brief Read a message from the COM port.
	
		The function reads data from the embedded CMT1s object. The data is then converted
		into a Message object.
		If an error occurred, a NULL pointer is returned and the error code can be
		retrieved with getLastError().
	*/
	Message* readMessage(void);
	
	/*! \brief Send a message over the COM port.
	
		The function attempts to write the message over the connected COM port.
		\param msg	The message to send.
		\param destroyMessageOnSuccess If this value is set to true (default), the
					Message object that was passed to the function is destroyed IF the
					function completed succesfully.
	*/
	CMTResult writeMessage(Message* msg, const bool destroyMessageOnSuccess = true);
	
	/*! \brief Wait for a message to arrive.
	
		The function waits for a message to arrive or until a timeout occurs.
		If the msgId parameter is set to a value other than 0, the function will wait
		until a message has arrived with that particular msgId.
		
		\note msgId 0 is the ReqDeviceID MID, but that is an outgoing only message. It is
		illogical to wait for a message that will never be sent to the host. So 0 is a
		safe value for the 'all' messages option.
	*/
	CMTResult waitForMessage(const unsigned char msgId = 0);

	//! \brief Return whether the communication port is open or not.
	bool isOpen (void) const;

	/*! \brief Open a communication channel to the given COM port number.
	
		This function is passed through to the CMT1s object.
	*/
	CMTResult open (const unsigned char portNumber,
					const unsigned long baudRate = CMT1S_DEFAULT_BAUD_RATE);

	/*! \brief Set the sizes of the inbound and outbound data buffers.
	
		This function is passed through to the CMT1s object.
	*/
	CMTResult setReadBufferSize (const unsigned long readBufSize);

	/*! \brief Set the default timeout value to use in blocking operations.

		This function is passed through to the CMT1s object.
	*/
	CMTResult setTimeout (const unsigned long ms);
};

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// CMT2f  /////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief The mid-level file communication class.

	This class handles log-files, parsing the XML-style data.
	It also handles writing to log files.
	The class uses CMT level 1, but does not inherit from it. If software needs to access 
	the level 1 component, it needs to be done through the cmt1f() function.
*/
class CMT2f {
private:
	//! This object cannot be copied, so this function is not implemented.
	CMT2f(const CMT2f& ref);

protected:
	CMTResult m_lastResult;			//!< The last result of an operation

	//! The CMT1f object that is used for the low-level operations
	CMT1f m_cmt1f;

	unsigned long m_dataPos;	//!< The position of the current data item.
	unsigned long m_configPos;	//!< The position of the configuration data.

//	unsigned char m_busId;

public:
	//! Default constructor, initializes all members to their default values.
	CMT2f();

	//! Destructor.
	~CMT2f();

	//! Close the file.
	CMTResult close (void);

	//! Return the error code of the last operation.
	CMTResult getLastError(void);

	/*! \brief Return a reference to the embedded CMT1f object.
	
		Any manipulation of the object should be done through CMT2f. CMT2f integrity is 
		not guaranteed if the CMT1f object is manipulated directly.
	*/
	CMT1f* cmt1f(void);

	/*! \brief Retrieve the filename that was last succesfully opened.

		This function is passed through to the embedded CMT1f object.
	*/
	CMTResult getName (unsigned char* filename) const;

	//! Return whether the file is open or not.
	bool isOpen (void) const;

	/*! \brief Open a file and read configuration data.
	
		The function opens a log file and scans it for configuration data and measurement
		data.
		If necessary, the configuration data is interpreted as far as necessary for
		creating correct data messages.
	*/
	CMTResult open (const char* filename);

	/*! \brief Return a setting as stored in the configuration part of the file.
		
		The function places the single byte setting with the appropriate name into the
		output buffer.
	*/
	CMTResult getSettingByte(const unsigned char* setting, unsigned char& output);

	/*! \brief Return a setting as stored in the configuration part of the file.
		
		The function places the single short setting with the appropriate name into the
		output buffer.
	*/
	CMTResult getSettingShort(const unsigned char* setting, unsigned short& output);

	/*! \brief Return a setting as stored in the configuration part of the file.
		
		The function places the single long setting with the appropriate name into the
		output buffer.
	*/
	CMTResult getSettingLong(const unsigned char* setting, unsigned long& output);
	
	/*! \brief Return a setting as stored in the configuration part of the file.
		
		The function places the float setting with the appropriate name into the output
		buffer.
	*/
	CMTResult getSettingFloat(const unsigned char* setting, float& output);
	
	/*! \brief Return a setting as stored in the configuration part of the file.
		
		The function places the setting with the appropriate name into the output buffer.
		The buffer is expected to be large enough to hold the data.
	*/
	CMTResult getSettingString(const unsigned char* setting, unsigned char* output);

	/*! \brief Restart the data reading.
	
		The function searches for the first (or nth) data item in the file so it can start
		reading from that position for any future requests.
		
		\param skip	The nr of items that will be skipped (from the start).
	*/
	CMTResult restartData(unsigned long skip = 0);

	/*! \brief Read data message from the file.
	
		The function reads data from the file and places it into a Message structure.
		The appropriate envelope is automatically generated by the function.
		
		If no data could be read, a NULL pointer is returned and the error can be
		retrieved with getLastError().
	*/
	Message* readDataMessage (void);

	/*! \brief Write data message to the file.
	
		The function writes a new data message to the end of the logfile. If the data was
		succesfully written and the destroyMessageOnSuccess parameter is true (default)
		the Message object is automatically destroyed.
	*/
	CMTResult writeDataMessage (Message* data, bool destroyMessageOnSuccess = true);

	/*! \brief Write configuration data to the file.
	
		The function writes the given configuration data to the logfile.
	*/
	CMTResult writeConfiguration (const DeviceConfiguration* config);
};

} // end of xsens namespace

#endif
